@@ -1,19 +1,20 @@
 /*
  * SPDX-FileCopyrightText: 2020 Arjen Hiemstra <ahiemstra@heimr.nl>
- * 
+ *
  * SPDX-License-Identifier: LGPL-2.1-only OR LGPL-3.0-only OR LicenseRef-KDE-Accepted-LGPL
  */
 
 #include "toolbarlayout.h"
 
-#include <unordered_map>
 #include <cmath>
+#include <unordered_map>
 
+#include <QDeadlineTimer>
 #include <QQmlComponent>
 #include <QTimer>
-#include <QDeadlineTimer>
 
 #include "enums.h"
+#include "loggingcategory.h"
 #include "toolbarlayoutdelegate.h"
 
 ToolBarLayoutAttached::ToolBarLayoutAttached(QObject *parent)
@@ -34,19 +35,22 @@ void ToolBarLayoutAttached::setAction(QObject *action)
 class ToolBarLayout::Private
 {
 public:
-    Private(ToolBarLayout *qq) : q(qq) { }
+    Private(ToolBarLayout *qq)
+        : q(qq)
+    {
+    }
 
     void performLayout();
-    QVector<ToolBarLayoutDelegate*> createDelegates();
+    QVector<ToolBarLayoutDelegate *> createDelegates();
     ToolBarLayoutDelegate *createDelegate(QObject *action);
     qreal layoutStart(qreal layoutWidth);
     void maybeHideDelegate(int index, qreal &currentWidth, qreal totalWidth);
 
     ToolBarLayout *q;
 
-    QVector<QObject*> actions;
+    QVector<QObject *> actions;
     ActionsProperty actionsProperty;
-    QList<QObject*> hiddenActions;
+    QList<QObject *> hiddenActions;
     QQmlComponent *fullDelegate = nullptr;
     QQmlComponent *iconDelegate = nullptr;
     QQmlComponent *moreButton = nullptr;
@@ -59,26 +63,32 @@ public:
     bool completed = false;
     bool layoutQueued = false;
     bool actionsChanged = false;
-    std::unordered_map<QObject*, std::unique_ptr<ToolBarLayoutDelegate>> delegates;
-    QVector<ToolBarLayoutDelegate*> sortedDelegates;
+    std::unordered_map<QObject *, std::unique_ptr<ToolBarLayoutDelegate>> delegates;
+    QVector<ToolBarLayoutDelegate *> sortedDelegates;
     QQuickItem *moreButtonInstance = nullptr;
     ToolBarDelegateIncubator *moreButtonIncubator = nullptr;
+    bool shouldShowMoreButton = false;
     int firstHiddenIndex = -1;
 
-    QVector<QObject*> removedActions;
+    QVector<QObject *> removedActions;
     QTimer *removalTimer = nullptr;
 
     QElapsedTimer performanceTimer;
 
     static void appendAction(ToolBarLayout::ActionsProperty *list, QObject *action);
+#if QT_VERSION < QT_VERSION_CHECK(6, 0, 0)
     static int actionCount(ToolBarLayout::ActionsProperty *list);
     static QObject *action(ToolBarLayout::ActionsProperty *list, int index);
+#else
+    static qsizetype actionCount(ToolBarLayout::ActionsProperty *list);
+    static QObject *action(ToolBarLayout::ActionsProperty *list, qsizetype index);
+#endif
     static void clearActions(ToolBarLayout::ActionsProperty *list);
 };
 
 ToolBarLayout::ToolBarLayout(QQuickItem *parent)
     : QQuickItem(parent)
-    , d(new Private{this})
+    , d(std::make_unique<Private>(this))
 {
     d->actionsProperty = ActionsProperty(this, this, Private::appendAction, Private::actionCount, Private::action, Private::clearActions);
 
@@ -89,7 +99,7 @@ ToolBarLayout::ToolBarLayout(QQuickItem *parent)
     d->removalTimer->setInterval(1000);
     d->removalTimer->setSingleShot(true);
     connect(d->removalTimer, &QTimer::timeout, this, [this]() {
-        for (auto action : qAsConst(d->removedActions)) {
+        for (auto action : std::as_const(d->removedActions)) {
             if (!d->actions.contains(action)) {
                 d->delegates.erase(action);
             }
@@ -107,7 +117,7 @@ ToolBarLayout::ActionsProperty ToolBarLayout::actionsProperty() const
     return d->actionsProperty;
 }
 
-void ToolBarLayout::addAction(QObject* action)
+void ToolBarLayout::addAction(QObject *action)
 {
     d->actions.append(action);
     d->actionsChanged = true;
@@ -127,7 +137,7 @@ void ToolBarLayout::addAction(QObject* action)
     relayout();
 }
 
-void ToolBarLayout::removeAction(QObject* action)
+void ToolBarLayout::removeAction(QObject *action)
 {
     auto itr = d->delegates.find(action);
     if (itr != d->delegates.end()) {
@@ -144,7 +154,7 @@ void ToolBarLayout::removeAction(QObject* action)
 
 void ToolBarLayout::clearActions()
 {
-    for (auto action : qAsConst(d->actions)) {
+    for (auto action : std::as_const(d->actions)) {
         auto itr = d->delegates.find(action);
         if (itr != d->delegates.end()) {
             itr->second->hide();
@@ -158,12 +168,12 @@ void ToolBarLayout::clearActions()
     relayout();
 }
 
-QList<QObject*> ToolBarLayout::hiddenActions() const
+QList<QObject *> ToolBarLayout::hiddenActions() const
 {
     return d->hiddenActions;
 }
 
-QQmlComponent * ToolBarLayout::fullDelegate() const
+QQmlComponent *ToolBarLayout::fullDelegate() const
 {
     return d->fullDelegate;
 }
@@ -180,7 +190,7 @@ void ToolBarLayout::setFullDelegate(QQmlComponent *newFullDelegate)
     Q_EMIT fullDelegateChanged();
 }
 
-QQmlComponent * ToolBarLayout::iconDelegate() const
+QQmlComponent *ToolBarLayout::iconDelegate() const
 {
     return d->iconDelegate;
 }
@@ -197,7 +207,6 @@ void ToolBarLayout::setIconDelegate(QQmlComponent *newIconDelegate)
     Q_EMIT iconDelegateChanged();
 }
 
-
 QQmlComponent *ToolBarLayout::moreButton() const
 {
     return d->moreButton;
@@ -306,15 +315,23 @@ void ToolBarLayout::componentComplete()
     relayout();
 }
 
-void ToolBarLayout::geometryChanged(const QRectF& newGeometry, const QRectF& oldGeometry)
+#if QT_VERSION < QT_VERSION_CHECK(6, 0, 0)
+void ToolBarLayout::geometryChanged(const QRectF &newGeometry, const QRectF &oldGeometry)
+#else
+void ToolBarLayout::geometryChange(const QRectF &newGeometry, const QRectF &oldGeometry)
+#endif
 {
     if (newGeometry != oldGeometry) {
         relayout();
     }
+#if QT_VERSION < QT_VERSION_CHECK(6, 0, 0)
     QQuickItem::geometryChanged(newGeometry, oldGeometry);
+#else
+    QQuickItem::geometryChange(newGeometry, oldGeometry);
+#endif
 }
 
-void ToolBarLayout::itemChange(QQuickItem::ItemChange change, const QQuickItem::ItemChangeData& data)
+void ToolBarLayout::itemChange(QQuickItem::ItemChange change, const QQuickItem::ItemChangeData &data)
 {
     if (change == ItemVisibleHasChanged || change == ItemSceneChange) {
         relayout();
@@ -330,7 +347,7 @@ void ToolBarLayout::updatePolish()
 void ToolBarLayout::Private::performLayout()
 {
     if (!fullDelegate || !iconDelegate || !moreButton) {
-        qWarning() << "ToolBarLayout: Unable to layout, required properties are not set";
+        qCWarning(KirigamiLog) << "ToolBarLayout: Unable to layout, required properties are not set";
         return;
     }
 
@@ -345,7 +362,7 @@ void ToolBarLayout::Private::performLayout()
 
     sortedDelegates = createDelegates();
 
-    bool ready = std::all_of(delegates.cbegin(), delegates.cend(), [](const std::pair<QObject* const, std::unique_ptr<ToolBarLayoutDelegate>> &entry) {
+    bool ready = std::all_of(delegates.cbegin(), delegates.cend(), [](const std::pair<QObject *const, std::unique_ptr<ToolBarLayoutDelegate>> &entry) {
         return entry.second->isReady();
     });
     if (!ready || !moreButtonInstance) {
@@ -358,7 +375,7 @@ void ToolBarLayout::Private::performLayout()
     // First, calculate the total width and maximum height of all delegates.
     // This will be used to determine which actions to show, which ones to
     // collapse to icon-only etc.
-    for (auto entry : qAsConst(sortedDelegates)) {
+    for (auto entry : std::as_const(sortedDelegates)) {
         if (!entry->isActionVisible()) {
             entry->hide();
             continue;
@@ -424,24 +441,38 @@ void ToolBarLayout::Private::performLayout()
             moreButtonInstance->setX(0.0);
         }
 
-        if (heightMode == AlwaysFill || (heightMode == ConstrainIfLarger && moreButtonInstance->height() > maxHeight)) {
+        if (heightMode == AlwaysFill) {
             moreButtonInstance->setHeight(q->height());
+        } else if (heightMode == ConstrainIfLarger) {
+            if (moreButtonInstance->implicitHeight() > maxHeight) {
+                moreButtonInstance->setHeight(maxHeight);
+            } else {
+                moreButtonInstance->setHeight(moreButtonInstance->implicitHeight());
+            }
         }
 
         moreButtonInstance->setY(qRound((maxHeight - moreButtonInstance->height()) / 2.0));
+        shouldShowMoreButton = true;
         moreButtonInstance->setVisible(true);
     } else {
+        shouldShowMoreButton = false;
         moreButtonInstance->setVisible(false);
     }
 
     qreal currentX = layoutStart(visibleActionsWidth);
-    for (auto entry : qAsConst(sortedDelegates)) {
+    for (auto entry : std::as_const(sortedDelegates)) {
         if (!entry->isVisible()) {
             continue;
         }
 
-        if (heightMode == AlwaysFill || (heightMode == ConstrainIfLarger && entry->height() > maxHeight)) {
+        if (heightMode == AlwaysFill) {
             entry->setHeight(q->height());
+        } else if (heightMode == ConstrainIfLarger) {
+            if (entry->implicitHeight() > maxHeight) {
+                entry->setHeight(maxHeight);
+            } else {
+                entry->setHeight(entry->implicitHeight());
+            }
         }
 
         qreal y = qRound((maxHeight - entry->height()) / 2.0);
@@ -480,10 +511,10 @@ void ToolBarLayout::Private::performLayout()
     sortedDelegates.clear();
 }
 
-QVector<ToolBarLayoutDelegate*> ToolBarLayout::Private::createDelegates()
+QVector<ToolBarLayoutDelegate *> ToolBarLayout::Private::createDelegates()
 {
-    QVector<ToolBarLayoutDelegate*> result;
-    for (auto action : qAsConst(actions)) {
+    QVector<ToolBarLayoutDelegate *> result;
+    for (auto action : std::as_const(actions)) {
         if (delegates.find(action) != delegates.end()) {
             result.append(delegates.at(action).get());
         } else {
@@ -500,10 +531,13 @@ QVector<ToolBarLayoutDelegate*> ToolBarLayout::Private::createDelegates()
         moreButtonIncubator->setStateCallback([this](QQuickItem *item) {
             item->setParentItem(q);
         });
-        moreButtonIncubator->setCompletedCallback([this](ToolBarDelegateIncubator* incubator) {
-            moreButtonInstance = qobject_cast<QQuickItem*>(incubator->object());
+        moreButtonIncubator->setCompletedCallback([this](ToolBarDelegateIncubator *incubator) {
+            moreButtonInstance = qobject_cast<QQuickItem *>(incubator->object());
             moreButtonInstance->setVisible(false);
 
+            connect(moreButtonInstance, &QQuickItem::visibleChanged, q, [this]() {
+                moreButtonInstance->setVisible(shouldShowMoreButton);
+            });
             connect(moreButtonInstance, &QQuickItem::widthChanged, q, [this]() {
                 Q_EMIT q->minimumWidthChanged();
             });
@@ -521,12 +555,12 @@ QVector<ToolBarLayoutDelegate*> ToolBarLayout::Private::createDelegates()
     return result;
 }
 
-ToolBarLayoutDelegate *ToolBarLayout::Private::createDelegate(QObject* action)
+ToolBarLayoutDelegate *ToolBarLayout::Private::createDelegate(QObject *action)
 {
     QQmlComponent *fullComponent = nullptr;
     auto displayComponent = action->property("displayComponent");
     if (displayComponent.isValid()) {
-        fullComponent = displayComponent.value<QQmlComponent*>();
+        fullComponent = displayComponent.value<QQmlComponent *>();
     }
 
     if (!fullComponent) {
@@ -537,7 +571,7 @@ ToolBarLayoutDelegate *ToolBarLayout::Private::createDelegate(QObject* action)
     result->setAction(action);
     result->createItems(fullComponent, iconDelegate, [this, action](QQuickItem *newItem) {
         newItem->setParentItem(q);
-        auto attached = static_cast<ToolBarLayoutAttached*>(qmlAttachedPropertiesObject<ToolBarLayout>(newItem, true));
+        auto attached = static_cast<ToolBarLayoutAttached *>(qmlAttachedPropertiesObject<ToolBarLayout>(newItem, true));
         attached->setAction(action);
     });
 
@@ -651,16 +685,24 @@ void ToolBarLayout::Private::maybeHideDelegate(int index, qreal &currentWidth, q
 
 void ToolBarLayout::Private::appendAction(ToolBarLayout::ActionsProperty *list, QObject *action)
 {
-    auto layout = reinterpret_cast<ToolBarLayout*>(list->data);
+    auto layout = reinterpret_cast<ToolBarLayout *>(list->data);
     layout->addAction(action);
 }
 
+#if QT_VERSION < QT_VERSION_CHECK(6, 0, 0)
 int ToolBarLayout::Private::actionCount(ToolBarLayout::ActionsProperty *list)
+#else
+qsizetype ToolBarLayout::Private::actionCount(ToolBarLayout::ActionsProperty *list)
+#endif
 {
-    return reinterpret_cast<ToolBarLayout*>(list->data)->d->actions.count();
+    return reinterpret_cast<ToolBarLayout *>(list->data)->d->actions.count();
 }
 
+#if QT_VERSION < QT_VERSION_CHECK(6, 0, 0)
 QObject *ToolBarLayout::Private::action(ToolBarLayout::ActionsProperty *list, int index)
+#else
+QObject *ToolBarLayout::Private::action(ToolBarLayout::ActionsProperty *list, qsizetype index)
+#endif
 {
     return reinterpret_cast<ToolBarLayout *>(list->data)->d->actions.at(index);
 }
